/*
 * Copyright (c) 2002, Regents of the University of California. All rights reserved.
 * 
 * ===================================================================
 * The Apache Software License, Version 1.1
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowledgment:
 *       "This product includes software developed by
 *        the Institute for Software Research at  
 *        University of California, Irvine"
 *    Alternately, this acknowledgment may appear in the software
 *    itself, if and wherever such third-party acknowledgments normally
 *    appear.
 * 
 * 4. The names "YANCEES", "Institute for Software Research" and
 *    "University of California" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    rsilvafi@ics.uci.edu or redmiles@ics.uci.edu.
 * 
 * 5. Products derived from this software may not be called
 *    "YANCEES", nor may "YANCEES" appear in their name, without
 *    prior written permission of the University of California Regents.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL WEBCOLLAB.COM OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 */
 package edu.uci.isr.yancees;

/**
 * <p>Title: Yancees Notification Server</p>
 * <p>Description: Yet ANother Configurable Extensible Event Service</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: School of Information and Computer Science - University of California, Irvine</p>
 * @author Roberto Silveira Silva Filho
 * @version 1.0
 */

import edu.uci.isr.yancees.core.ProtocolManager;

/**
 * This class deals with generic messages, that are parserd by installed protocol
 * plug-ins. For example, mobility, authentication and so on.
 * 
 * This class stays behind the YanceesServer class 
 *
 */
public class ProtocolFacade
      extends AdministrationFacade {

   private ProtocolManager protocolManager = null;

   // single instance of this class
   static ProtocolFacade myInstance;

   // can only be accessed by the YancessAPI or anyone else inside yancess package
   static protected ProtocolFacade getInstance() {
      if (myInstance == null) {
         myInstance = new ProtocolFacade();
      }
      return myInstance;
   }


   // private here prevents direct instantiation
   private ProtocolFacade() {
   }

   /**
    * Creates a new ProtocolSession which is initialized according to the message provided.
    * The messages of this session need now to be provided using the created ProtocolSession
    * object. The responses, specific to this protocol, are directed to the provided ProtocolListenerInterface.
    * the data, the events collected or generated by producers, are diverget to the SubscriberInterface.
    *
    * @param msg is the initialization message of the session
    * @param pl is the interface to receive the responses to the messages of the new session (control events)
    * @param si is the subscriber interface, to receive data events.
    * @return a ProtocolSession that controls the created context.
    * @throws YanceesException if the protocol is not recognized as valid.
    */
   public ProtocolSession createNewSession(MessageInterface msg,
                                           SubscriberInterface si) throws
         YanceesException {

      ((GenericMessage) msg).setDateReceivedInServer(new java.util.Date());
      ProtocolSession newSession = new ProtocolSession(si, protocolManager);
      // uses the message to instantiate a new plug-in, which will be active until termination
      newSession.initialize(msg);

      return newSession;
   }

   /**
    * Used by the Yancees API to initialize this API.
    * @param pm is the instance of the protocol manager to be used by this component
    */
   protected void setProtocolManager(ProtocolManager pm) {
      protocolManager = pm;
   }

}