/*
 * Copyright (c) 2002, Regents of the University of California. All rights reserved.
 * 
 * ===================================================================
 * The Apache Software License, Version 1.1
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowledgment:
 *       "This product includes software developed by
 *        the Institute for Software Research at  
 *        University of California, Irvine"
 *    Alternately, this acknowledgment may appear in the software
 *    itself, if and wherever such third-party acknowledgments normally
 *    appear.
 * 
 * 4. The names "YANCEES", "Institute for Software Research" and
 *    "University of California" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    rsilvafi@ics.uci.edu or redmiles@ics.uci.edu.
 * 
 * 5. Products derived from this software may not be called
 *    "YANCEES", nor may "YANCEES" appear in their name, without
 *    prior written permission of the University of California Regents.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL WEBCOLLAB.COM OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 */
 package edu.uci.isr.yancees.server.rmi;

import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.rmi.server.Unreferenced;

import edu.uci.isr.yancees.MessageInterface;
import edu.uci.isr.yancees.ProtocolSession;
import edu.uci.isr.yancees.YanceesException;

/**
 * <p>Title: Yancees Notification Server</p>
 * <p>Description: Yet ANother Configurable Extensible Event Service</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: School of Information and Computer Science - University of California, Irvine</p>
 * @author Roberto Silveira Silva Filho
 * @version 1.0
 */

/**
 * This class works as a remote wrapper to the local protocolSession
 * It is sent as a callback interface to the yancees clients
 */
public class RemoteProtocolSessionImplementation
      extends UnicastRemoteObject
      implements RemoteProtocolSessionInterface, Unreferenced {

   private ProtocolSession protocolSession;
   private boolean print = edu.uci.isr.yancees.YanceesFacade.PRINT_DEBUG;

   public RemoteProtocolSessionImplementation(ProtocolSession session)
         throws RemoteException {
      super();
      protocolSession = session;

   }

   public void sendMessage(MessageInterface message)
         throws YanceesException, RemoteException {
      protocolSession.sendMessage(message);
   }

   /**
    * Gracefully terminates the current session
    * @param message is a finalization message to be sent to the plug-in(s)
    * created to handle this session
    * @throws YanceesException in case of parsing or message errors
    * @throws RemoteException in case of connection errors
    */
   public void terminate(MessageInterface message)
         throws YanceesException, RemoteException {
      protocolSession.terminate(message);
   }

   /**
    * Gracefully terminates the current session
    * @throws YanceesException in case of parsing or message errors
    * @throws RemoteException in case of connection errors
    */
   public void terminate()
         throws YanceesException, RemoteException {
      protocolSession.terminate();
   }

   /**
    * Implements the Unreferenced interface.
    * This class is invoked by the Distributed Garbage Collector when the RMI reference
    * counter for this object becomes 0 or when there is no communication for more than
    * the lease period, which default timeout is 10 minutes.
    *
    * This method is used to release any resource allocated to this record which,
    * in this case, is the protocol plug-in.
    */
   public void unreferenced() {
      if (print)
         System.out.println(
               "RemoteProtocolSessionImplementation: terminating the plugin session...");

      try {
         terminate();
      } catch (RemoteException ex) {
         System.out.println(ex);
      } catch (YanceesException ex) {
         System.out.println(ex);
      }

   }

}